/* -*-c-*- */

/* ---------------------------- included header files ---------------------- */

#include <config.h>

#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <dirent.h>

#include "features/FILE_PERMISSIONS"

/* ---------------------------- local definitions -------------------------- */

#define N_PATH_TEST_FUNCS 6
/* for debugging this test program, set it to 1 */
#define GATEG_TEST_DEBUG 1

#define GATEG_DEBUG
#define MAX_TESTDIR_LEN 100

#define TFG_M0755 (GATEG_S_IRWXU | GATEG_S_IRGRP | GATEG_S_IXGRP | \
	GATEG_S_IROTH | GATEG_S_IXOTH)
#define TFG_M0775 (TFG_M0755 | GATEG_S_IWGRP)
#define TFG_M0757 (TFG_M0755 | GATEG_S_IWOTH)
#define TFG_M0077 (GATEG_S_IRWXG | GATEG_S_IRWXO)
#define TFG_M0777 (TFG_M0077 | GATEG_S_IRWXU)
#define TFG_M01777 (GATEG_S_ISVTX | TFG_M0755 | TFG_M0077)
#define TFG_M01500 (GATEG_S_ISVTX | GATEG_S_IRUSR | GATEG_S_IXUSR)
#define TFG_M01522 (TFG_M01500 | GATEG_S_IWGRP | GATEG_S_IWOTH)
#define TFG_M06755 (GATEG_S_ISUID | GATEG_S_ISGID | TFG_M0755)
#define TFG_M07755 (TFG_M06755 | GATEG_S_ISVTX)

/* ---------------------------- included code files ------------------------ */

#include "fileguardian.c"

/* ---------------------------- feature tests ------------------------------ */

/* ---------------------------- local macros ------------------------------- */

/* Aborts if the called code does not call abort() and vice versa.
 * This can be written only as a macro because it takes a piece of code as an
 * argument.
 */
#define ASSERT_ABORT(code) \
	do \
	{ \
		if (setjmp(__gateg_debug_jmp_buf) == 0) \
		{ \
			do { code; } while (0); \
			/* If we get here, longjmp was never called, */ \
			/* i.e. the assertion failed */ \
			abort(); \
		} \
		/* code triggered an abort(), ok */ \
	} while (0)

/* ---------------------------- imports ------------------------------------ */

/* ---------------------------- local types -------------------------------- */

typedef int (*test_func_t)(void);

typedef struct
{
	test_func_t t_func;
	const char *t_name;
	const char *t_msg;
	int t_resulttype;
} test_table_entry_t;

enum
{
	RET_OK = 0,
	/* errors */
	RET_FAILED = 255,
	RET_SKIPPED = 254,
	RET_EOT = -99
};

typedef struct
{
	uid_t d_owner[4];
	mode_t d_mode[4];
	mode_t d_mode_forbidden;
	int d_start_dir_depth;
	int d_do_ignore_sticky_bit;
	int d_result;
} dir_test_table_entry_t;

typedef struct
{
	const char *p_path;
	const char *p_path_nofuzz;
	int p_result[N_PATH_TEST_FUNCS];
} path_test_table_entry_t;

enum
{
	FUNC__SOWM,
	FUNC_SOWM,
	FUNC_SO,
	FUNC_CDSWM,
	FUNC_CCS,
	FUNC_STWN
};

/* ---------------------------- local variables ---------------------------- */

typedef int (*path_test_func_t)(const char *path);

static path_test_func_t path_test_funcs[N_PATH_TEST_FUNCS] =
{
	fileg_is_absolute_path,
	fileg_is_relative_path,
	fileg_is_local_relative_path,
	fileg_is_path,
	fileg_is_filename,
	(path_test_func_t)(fileg_remove_fuzz_from_path)
};

static int have_uids = 0;

static uid_t uids[3] =
{
	0,
	(uid_t)-1,
	(uid_t)-2
};

static char test_dir[MAX_TESTDIR_LEN + 2];
static char test_dir2[MAX_TESTDIR_LEN + 4];
static char test_dir3[MAX_TESTDIR_LEN + 6];
static char test_dir4[MAX_TESTDIR_LEN + 8];
static char *tdirs[5] =
{
	"/",
	test_dir,
	test_dir2,
	test_dir3,
	test_dir4
};

/* ---------------------------- local functions ---------------------------- */

static void __usage(int argc, char **argv)
{
	fprintf(stderr, "usage: %s uid1 uid2 dirname\n", argv[0]);
	fprintf(stderr, "usage: %s -n\n", argv[0]);
	fprintf(stderr, "  uid1   : numeric user id for testing purposes\n");
	fprintf(stderr, "  uid2   : numeric user id for testing purposes\n");
	fprintf(stderr, "  dirname: name of directory to create below root\n");
	fprintf(stderr, "           (must begin with /, e.g. /fgtestdir)\n");

	return;
}

static int __parse_options(int argc, char **argv)
{

	char *dir;
	char *s;
	int rc;
	int i;

	if (argc < 2)
	{
		__usage(argc, argv);

		return 1;
	}
	if (
		argc == 2 &&
		argv[1][0] == '-' && argv[1][1] == 'n' && argv[1][2] == 0)
	{
		/*!!!*/
		return 0;
	}
	if (argc < 4)
	{
		__usage(argc, argv);

		return 1;
	}
	rc = sscanf(argv[1], "%d", &i);
	if (rc != 1)
	{
		__usage(argc, argv);

		return 1;
	}
	uids[1] = (uid_t)i;
	rc = sscanf(argv[2], "%d", &i);
	if (rc != 1)
	{
		__usage(argc, argv);

		return 1;
	}
	uids[2] = (uid_t)i;
	if (uids[1] == uids[2] || uids[1] <= 0 || uids[2] <= 0)
	{
		fprintf(stderr, "uid1 and uid2 must be different and > 0\n");

		return 1;
	}
	dir = argv[3];
	if (dir == NULL || *dir != '/')
	{
		fprintf(stderr, "dirname must begin with '/'\n");

		return 1;
	}
	for (s = dir + 1; *s != 0; s++)
	{
		if (*s == '/')
		{
			fprintf(stderr, "dirname must contain a single '/'\n");

			return 1;
		}
	}
	if (s - dir > MAX_TESTDIR_LEN + 2)
	{
		fprintf(
			stderr, "dirname not be longer than %d characters\n",
			MAX_TESTDIR_LEN);

		return 1;
	}
	for (i = 0; i < MAX_TESTDIR_LEN + 1 && dir[i] != 0; i++)
	{
		tdirs[1][i] = dir[i];
		tdirs[2][i] = dir[i];
		tdirs[3][i] = dir[i];
		tdirs[4][i] = dir[i];
	}
	tdirs[1][i] = 0;
	tdirs[2][i] = '/';
	tdirs[3][i] = '/';
	tdirs[4][i] = '/';
	i++;
	tdirs[2][i] = '2';
	tdirs[3][i] = '2';
	tdirs[4][i] = '2';
	i++;
	tdirs[2][i] = 0;
	tdirs[3][i] = '/';
	tdirs[4][i] = '/';
	i++;
	tdirs[3][i] = '3';
	tdirs[4][i] = '3';
	i++;
	tdirs[3][i] = 0;
	tdirs[4][i] = '/';
	i++;
	tdirs[4][i] = '4';
	i++;
	tdirs[4][i] = 0;
	have_uids = 1;

	return 0;
}

/* ---------------------------- local variables ---------------------------- */

static dir_test_table_entry_t _sowm_test_table[] =
{
	/* root owner */
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 0, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 1, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 2, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 3, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 0
	},

	/* same owner */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 0, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 1, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 2, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 3, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 0
	},

	/* other owner */
	{
		{ 1, 1, 1, 2 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 2, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 2, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 2, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* wrong mode */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0775 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0775, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0775, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0775, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0757, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* wrong owner */
	{
		{ 1, 1, 1, 2 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 2, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 2, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 2, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* modes tests */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M06755, TFG_M06755, TFG_M06755, TFG_M06755 },
		00000, 4, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M07755, TFG_M07755, TFG_M07755, TFG_M07755 },
		00000, 4, 1, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M01522, TFG_M01522, TFG_M01522, TFG_M01522 },
		00000, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M01522, TFG_M01522, TFG_M01522, TFG_M01522 },
		00000, 4, 1, 0
	},

	{ { 0, 0, 0, 0 }, { 0, 0, 0, 0 }, 0, 0, 0, RET_EOT }
};

static dir_test_table_entry_t sowm_test_table[] =
{
	/* root owner */
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 0, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 1, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 2, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 3, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 0
	},

	/* same owner */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 0, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 1, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 2, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 3, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 0
	},

	/* other owner */
	{
		{ 1, 1, 1, 2 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 2, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 2, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 2, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* wrong mode */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0775 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0775, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0775, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0775, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0757, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* wrong owner */
	{
		{ 1, 1, 1, 2 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 2, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 2, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 2, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* modes tests */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M06755, TFG_M06755, TFG_M06755, TFG_M06755 },
		00000, 4, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M01522, TFG_M01522, TFG_M01522, TFG_M01522 },
		00000, 4, 0, 1
	},

	{ { 0, 0, 0, 0 }, { 0, 0, 0, 0 }, 0, 0, 0, RET_EOT }
};

static dir_test_table_entry_t so_test_table[] =
{
	/* root owner */
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 0, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 1, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 2, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 3, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 0
	},

	/* same owner */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 0, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 1, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 2, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 3, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 0
	},

	/* other owner */
	{
		{ 1, 1, 1, 2 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 2, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 2, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 2, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* wrong mode */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0775 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0775, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0775, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0775, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0757, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* wrong owner */
	{
		{ 1, 1, 1, 2 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 2, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 2, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 2, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* modes tests */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M06755, TFG_M06755, TFG_M06755, TFG_M06755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M01522, TFG_M01522, TFG_M01522, TFG_M01522 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	{ { 0, 0, 0, 0 }, { 0, 0, 0, 0 }, 0, 0, 0, RET_EOT }
};

static dir_test_table_entry_t cdswm_test_table[] =
{
	/* root owner */
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 0, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 1, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 2, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 3, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 0
	},

	/* same owner */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 0, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 1, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 2, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 3, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 0
	},

	/* other owner */
	{
		{ 1, 1, 1, 2 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 2, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 2, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 2, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* wrong mode */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0775 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0775, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0775, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0775, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0757, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* wrong owner */
	{
		{ 1, 1, 1, 2 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 2, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 2, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 2, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* modes tests */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M06755, TFG_M06755, TFG_M06755, TFG_M06755 },
		00000, 4, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M01522, TFG_M01522, TFG_M01522, TFG_M01522 },
		00000, 4, 0, 1
	},

	{ { 0, 0, 0, 0 }, { 0, 0, 0, 0 }, 0, 0, 0, RET_EOT }
};

static dir_test_table_entry_t ccs_test_table[] =
{
	/* root owner */
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 0, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 1, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 2, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 3, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 0
	},

	/* same owner */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 0, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 1, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 2, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 3, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 0
	},

	/* other owner */
	{
		{ 1, 1, 1, 2 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 2, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 2, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 2, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* wrong mode */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0775 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0775, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0755, TFG_M0775, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0775, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M0757, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* wrong owner */
	{
		{ 1, 1, 1, 2 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 1, 2, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 1, 2, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},
	{
		{ 2, 1, 1, 1 },
		{ TFG_M0755, TFG_M0755, TFG_M0755, TFG_M0755 },
		FILEG_DEFAULT_FORBIDDEN_DIR_MODE, 4, 0, 1
	},

	/* modes tests */
	{
		{ 1, 1, 1, 1 },
		{ TFG_M06755, TFG_M06755, TFG_M06755, TFG_M06755 },
		00000, 4, 0, 0
	},
	{
		{ 1, 1, 1, 1 },
		{ TFG_M01522, TFG_M01522, TFG_M01522, TFG_M01522 },
		00000, 4, 0, 1
	},

	{ { 0, 0, 0, 0 }, { 0, 0, 0, 0 }, 0, 0, 0, RET_EOT }
};

static dir_test_table_entry_t stwn_test_table[] =
{
	{
		{ 1, 0, 0, 0 },
		{ TFG_M0755, 0, 0, 0 },
		0, 1, 0, 0
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M0777, 0, 0, 0 },
		0, 1, 0, 1
	},
	{
		{ 0, 0, 0, 0 },
		{ TFG_M01777, 0, 0, 0 },
		0, 1, 0, 0
	},
	{ { 0, 0, 0, 0 }, { 0, 0, 0, 0 }, 0, 0, 0, RET_EOT }
};

static path_test_table_entry_t path_test_table[] =
{
	{
		".", ".", { 0, 1, 1, 1, 0, 0 }
	},
	{
		0, 0, { 0, 0, 0, 0, 0, 0 }
	},
	{
		"", "", { 0, 0, 0, 0, 0, 0 }
	},
	{
		"/", "/", { 1, 0, 0, 1, 0, 0 }
	},
	{
		"/x", "/x", { 1, 0, 0, 1, 0, 0 }
	},
	{
		"/xy", "/xy", { 1, 0, 0, 1, 0, 0 }
	},
	{
		"/xy/", "/xy", { 1, 0, 0, 1, 0, 1 }
	},
	{
		".", ".", { 0, 1, 1, 1, 0, 0 }
	},
	{
		"./", ".", { 0, 1, 1, 1, 0, 1 }
	},
	{
		"./x", "x", { 0, 1, 1, 1, 0, 1 }
	},
	{
		"..", "..", { 0, 1, 0, 1, 0, 0 }
	},
	{
		"../", "..", { 0, 1, 0, 1, 0, 1 }
	},
	{
		"../x", "../x", { 0, 1, 0, 1, 0, 0 }
	},
	{
		"x/y", "x/y", { 0, 1, 1, 1, 0, 0 }
	},
	{
		"/..", "/", { 1, 0, 0, 1, 0, 1 }
	},
	{
		"/.", "/", { 1, 0, 0, 1, 0, 1 }
	},
	{
		"/../x", "/x", { 1, 0, 0, 1, 0, 1 }
	},
	{
		"/../../.", "/", { 1, 0, 0, 1, 0, 1 }
	},
	{
		"/../", "/", { 1, 0, 0, 1, 0, 1 }
	},
	{
		"//", "/", { 1, 0, 0, 1, 0, 1 }
	},
	{
		"x//y", "x/y", { 0, 1, 1, 1, 0, 1 }
	},
	{
		"x/y//", "x/y", { 0, 1, 1, 1, 0, 1 }
	},
	{
		"x/./", "x", { 0, 1, 1, 1, 0, 1 }
	},
	{
		"/./", "/", { 1, 0, 0, 1, 0, 1 }
	},
	{
		"x/./y", "x/y", { 0, 1, 1, 1, 0, 1 }
	},
	{
		"x/.", "x", { 0, 1, 1, 1, 0, 1 }
	},
	{
		"xyz", "xyz", { 0, 0, 1, 0, 1, 0 }
	},
	{ 0, 0, { RET_EOT, RET_EOT, RET_EOT, RET_EOT, RET_EOT, RET_EOT } }
};

/* ---------------------------- test functions ----------------------------- */

static int __tdirs_set_properties(dir_test_table_entry_t *tabentry)
{
	int rc;
	int i;

	for (i = 0; i < 4; i++)
	{
		rc = chown(tdirs[i + 1], uids[tabentry->d_owner[i]], (gid_t)0);
		if (rc != 0)
		{
			fprintf(
				stderr,
				"    unable to change owner/group: %s\n",
				tdirs[i + 1]);

			return rc;
		}
		rc = chmod(tdirs[i + 1], tabentry->d_mode[i]);
		if (rc != 0)
		{
			fprintf(
				stderr,
				"    unable to change permissions: %s\n",
				tdirs[i + 1]);

			return rc;
		}
	}

	return 0;
}

static int __test_with_dir_table(dir_test_table_entry_t *table, int func)
{
	char name[L_tmpnam];
	FILE *f;
	DIR *d;
	int rc;
	int i;

	if (have_uids == 0)
	{
		fprintf(stderr, "    this test needs two uids to run\n");

		return RET_SKIPPED;
	}
	if (geteuid() != 0)
	{
		fprintf(stderr, "    this test can only run with uid 0\n");

		return RET_SKIPPED;
	}
	if (getegid() != 0)
	{
		fprintf(stderr, "    this test can only run with gid 0\n");

		return RET_SKIPPED;
	}
	for (i = 0, rc = 0; table[i].d_result != RET_EOT && rc == 0; i++)
	{
		fprintf(stderr, "    subtest %d\n", i);
		rc = __tdirs_set_properties(&(table[i]));
		if (rc != 0)
		{
			break;
		}
		rc = seteuid(uids[0]);
		if (rc != 0)
		{
			fprintf(stderr, "    unable to change to uid 0\n");

			break;
		}
		rc = seteuid(uids[1]);
		if (rc != 0)
		{
			fprintf(
				stderr, "    unable to change to uid %d\n",
				uids[1]);

			break;
		}
		/* check directory */
		switch (func)
		{
		case FUNC__SOWM:
			d = __fileg_safe_opendir_with_mode(
				tdirs[table[i].d_start_dir_depth],
				table[i].d_mode_forbidden,
				table[i].d_do_ignore_sticky_bit);
			rc = (d == 0);
			if (d != NULL)
			{
				closedir(d);
			}
			break;
		case FUNC_SOWM:
			d = fileg_safe_opendir_with_mode(
				tdirs[table[i].d_start_dir_depth],
				table[i].d_mode_forbidden);
			rc = (d == 0);
			if (d != NULL)
			{
				closedir(d);
			}
			break;
		case FUNC_SO:
			d = fileg_safe_opendir(
				tdirs[table[i].d_start_dir_depth]);
			rc = (d == 0);
			if (d != NULL)
			{
				closedir(d);
			}
			break;
		case FUNC_CDSWM:
			rc = fileg_check_dir_safety_with_mode(
				tdirs[table[i].d_start_dir_depth],
				table[i].d_mode_forbidden);
			rc =!!rc;
			break;
		case FUNC_CCS:
			rc = chdir(tdirs[table[i].d_start_dir_depth]);
			if (rc != 0)
			{
				fprintf(
					stderr, "    chdir failed: %s\n",
					tdirs[table[i].d_start_dir_depth]);
				abort();
				exit(255);
			}
			rc = fileg_check_cwd_safety();
			rc =!!rc;
			if (chdir("/") != 0)
			{
				fprintf(
					stderr, "    chdir failed: %s\n",
					tdirs[table[i].d_start_dir_depth]);
				abort();
				exit(255);
			}
			break;
		case FUNC_STWN:
			f = fileg_safe_tmpfile_with_name(name, tdirs[1]);
			rc = (f == 0);
			if (f != 0)
			{
				fclose(f);
				unlink(name);
			}
			break;
		default:
			abort();
			exit(255);
		}
		if (rc != !!(table[i].d_result))
		{
			return RET_FAILED;
		}
		/* clean up */
		rc = seteuid(uids[0]);
		if (rc != 0)
		{
			fprintf(stderr, "    unable to change to uid 0\n");

			break;
		}
	}
	if (rc != 0)
	{
		return RET_FAILED;
	}

	return RET_OK;
}

static int __test___safe_opendir_with_mode(void)
{
	int rc;

	rc = __test_with_dir_table(&(_sowm_test_table[0]), FUNC__SOWM);

	return rc;
}

static int __test_safe_opendir_with_mode(void)
{
	int rc;

	rc = __test_with_dir_table(&(sowm_test_table[0]), FUNC_SOWM);

	return rc;
}

static int __test_safe_opendir(void)
{
	int rc;

	rc = __test_with_dir_table(&(so_test_table[0]), FUNC_SO);

	return rc;
}

static int __test_check_dir_safety_with_mode(void)
{
	int rc;

	rc = __test_with_dir_table(&(cdswm_test_table[0]), FUNC_CDSWM);

	return rc;
}

static int __test_check_cwd_safety(void)
{
	int rc;

	rc = __test_with_dir_table(&(ccs_test_table[0]), FUNC_CCS);

	return rc;
}

static int __test_tempfiles(void)
{
	int rc;

	rc = __test_with_dir_table(&(stwn_test_table[0]), FUNC_STWN);

	return rc;
}

static int __test_path_tests(void)
{
	int rc;
	int i;
	int j;

	for (i = 0; path_test_table[i].p_result[0] != RET_EOT; i++)
	{
		const char *path;
		char *p;

		path = path_test_table[i].p_path;
		fprintf(
			stderr, "    subtest %d '%s'\n", i,
			(path == 0) ? "(null)" : path);
		if (path != 0)
		{
			const char *s;
			char *d;
			size_t len;

			len = __fileg_strlen(path);
			p = (char *)malloc(len + 1);
			if (p == 0)
			{
				fprintf(
					stderr,
					"    unable to allocate memory\n");
				return RET_FAILED;
			}
			for (s = path, d = p; *s != 0; s++, d++)
			{
				*d = *s;
			}
			*d = 0;
		}
		else
		{
			p = 0;
		}
		for (j = 0; j < N_PATH_TEST_FUNCS - 1; j++)
		{
			rc = path_test_funcs[j](path);
			if (rc != path_test_table[i].p_result[j])
			{
				fprintf(
					stderr,
					"    wrong rc %d in function %d\n",
					rc, j);
				return RET_FAILED;
			}
		}
		rc = path_test_funcs[j](p);
		if (rc != path_test_table[i].p_result[j])
		{
			fprintf(
				stderr,
				"    wrong rc %d in function %d '%s'\n",
				rc, j, p);
			return RET_FAILED;
		}
		if (p != 0)
		{
			if (
				__fileg_strlen(p) != __fileg_strlen(
					path_test_table[i].p_path_nofuzz))
			{
				fprintf(
					stderr, "    wrong result path '%s',"
					" expected '%s'\n", p,
					path_test_table[i].p_path_nofuzz);
				return RET_FAILED;
			}
			free(p);
		}
	}

	return RET_OK;
}

/* ---------------------------- local variables ---------------------------- */

static test_table_entry_t test_table[] =
{
	{ __test___safe_opendir_with_mode, "__safe_opendir_with_mode", 0,
	  RET_OK },
	{ __test_safe_opendir_with_mode, "safe_opendir_with_mode", 0, RET_OK },
	{ __test_safe_opendir, "safe_opendir", 0, RET_OK },
	{ __test_check_dir_safety_with_mode, "check_dir_safety_with_mode", 0,
	  RET_OK },
	{ __test_check_cwd_safety, "check_cwd_safety", 0, RET_OK },
	{ __test_tempfiles, "safe temporary files", 0, RET_OK },
	{ __test_path_tests, "test path test functions", 0, RET_OK },
	{ 0, 0, 0, 0 }
};

/* ---------------------------- local functions ---------------------------- */

static int __run_test(test_table_entry_t *test)
{
	pid_t child = 0;
	pid_t pid = 0;
	int status = 0;
	int rc;

	child = 0;
	pid = 0;
	status = 0;
	fprintf(stderr, "*** executing test %s\n", test->t_name);
	if (test->t_msg != 0)
	{
		fprintf(stderr, "    %s\n", test->t_msg);
	}
	if (GATEG_TEST_DEBUG == 0)
	{
		child = fork();
		switch (child)
		{
		case (pid_t)-1:
			perror("fork");
			abort();
		case 0:
			/* child */
			rc = test->t_func();
			exit(rc);
		default:
			/* parent */
			pid = wait(&status);
			rc = WEXITSTATUS(status);
			break;
		}
	}
	else
	{
		rc = test->t_func();
	}
	switch (rc)
	{
	case RET_OK:
		fprintf(stderr, "OK\n");
		break;
	case RET_SKIPPED:
		fprintf(stderr, "SKIPPED test\n");
		break;
	case RET_FAILED:
	default:
		fprintf(stderr, "FAILED with return code %d\n", rc);
		break;
	}

	return rc;
}

static void __setup_test_env(void)
{
	int rc;

	if (have_uids == 0)
	{
		return;
	}
	umask(TFG_M0077);
	rc = mkdir(tdirs[1], TFG_M0755);
	if (rc != 0)
	{
		perror("mkdir");
		fprintf(stderr, "failed to create directory '%s'\n", tdirs[1]);
		fprintf(
			stderr,
			"if it exists and is not empty, please remove it"
			" manually\n");
		exit(2);
	}
	rc = mkdir(tdirs[2], TFG_M0755);
	if (rc != 0)
	{
		perror("mkdir");
		fprintf(
			stderr, "failed to create directory '%s'\n",
			tdirs[2]);
		fprintf(
			stderr,
			"if it exists and is not empty, please remove it"
			" manually\n");
		exit(2);
	}
	rc = mkdir(tdirs[3], TFG_M0755);
	if (rc != 0)
	{
		perror("mkdir");
		fprintf(
			stderr, "failed to create directory '%s'\n",
			tdirs[3]);
		fprintf(
			stderr,
			"if it exists and is not empty, please remove it"
			" manually\n");
		exit(2);
	}
	rc = mkdir(tdirs[4], TFG_M0755);
	if (rc != 0)
	{
		perror("mkdir");
		fprintf(
			stderr, "failed to create directory '%s'\n",
			tdirs[4]);
		fprintf(
			stderr,
			"if it exists and is not empty, please remove it"
			" manually\n");
		exit(2);
	}
	umask(0000);

	return;
}

static void __cleanup_test_env(void)
{
	if (have_uids == 0)
	{
		return;
	}
	seteuid((uid_t)0);
	rmdir(tdirs[4]);
	rmdir(tdirs[3]);
	rmdir(tdirs[2]);
	rmdir(tdirs[1]);

	return;
}

/* ---------------------------- interface functions ------------------------ */

int main(int argc, char **argv)
{
	int tests_failed;
	int tests_skipped;
	int rc;
	int i;

	rc = __parse_options(argc, argv);
	if (rc != 0)
	{
		exit(1);
	}
	fprintf(
		stderr, "If this program aborts, there is a flaw in the"
		" tested code\n");
	__setup_test_env();
	tests_failed = 0;
	tests_skipped = 0;
	for (i = 0; test_table[i].t_func != 0; i++)
	{
		rc = __run_test(&test_table[i]);
		switch (rc)
		{
		case RET_FAILED:
			tests_failed++;
			break;
		case RET_SKIPPED:
			tests_skipped++;
			break;
		case RET_OK:
			break;
		default:
			fprintf(stderr, "\nERROR: invalid return code\n");
			abort();
			exit(1);
			break;
		}
	}
	__cleanup_test_env();
	if (tests_skipped != 0)
	{
		fprintf(
			stderr, "\nWARNING: %d tests have been skipped\n",
			tests_skipped);
	}
	if (tests_failed != 0)
	{
		fprintf(stderr, "\nERROR: at least one test has failed\n");

		return 1;
	}

	return 0;
}
