/* -*-c-*- */

/* ---------------------------- included header files ---------------------- */

#include <config.h>

#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <dirent.h>
#include <unistd.h>

/* ---------------------------- local definitions -------------------------- */

/* for debugging this test program, set it to 1 */
#define GATEG_TEST_DEBUG 0

#define GATEG_DEBUG
#define TEST_DIR "./testdir"
#define NONEXIST_DIR "./nonexist"
#define CORE_NAME "./core"
#define CORE_NAME2 "./testgg.core"
#define CORE_NAME3 "./testgg_cpp.core"
#define CORE_NAME4_PREFIX "./testgg"
#define CORE_NAME5_PREFIX "./testgg_cpp"

/* ---------------------------- included code files ------------------------ */

#include "gateguardian.c"

/* ---------------------------- feature tests ------------------------------ */

#ifdef HAVE_DIRFD
#undef HAVE_DIRFD
#define HAVE_DIRFD 1
#define GATEG_DIRFD(x) dirfd(x)
#else
#define HAVE_DIRFD 0
#define GATEG_DIRFD(x) 0
#endif
#ifdef GATEG_RLIMIT_CORE
#define TEST_CORE 1
#else
#define TEST_CORE 0
#endif

/* ---------------------------- local macros ------------------------------- */

/* Aborts if the called code does not call abort() and vice versa.
 * This can be written only as a macro because it takes a piece of code as an
 * argument.
 */
#define ASSERT_ABORT(code) \
	do \
	{ \
		if (setjmp(__gateg_debug_jmp_buf) == 0) \
		{ \
			do { code; } while (0); \
			/* If we get here, longjmp was never called, */ \
			/* i.e. the assertion failed */ \
			abort(); \
		} \
		/* code triggered an abort(), ok */ \
	} while (0)

/* ---------------------------- imports ------------------------------------ */

/* ---------------------------- local types -------------------------------- */

typedef int (*test_func_t)(void);

typedef struct
{
	test_func_t t_func;
	const char *t_name;
	const char *t_msg;
	int t_resulttype;
} test_table_entry_t;

enum
{
	RET_OK = 0,
	RET_EXPECT_CORE,
	RET_EXPECT_NO_CORE,
	/* errors */
	RET_FAILED = 255,
	RET_SKIPPED = 254,
	RET_SKIP_NO_CORE = 253
};

/* ---------------------------- local functions ---------------------------- */

static int __str_equal(const char *s1, const char *s2)
{
	if (s1 == 0)
	{
		return (s2 == 0);
	}
	for ( ; *s1 != 0 && *s1 == *s2; s1++, s2++)
	{
		/* nothing */
	}
	if (*s1 == 0 && *s2 == 0)
	{
		return 1;
	}

	return 0;
}

static int __pstat(DIR *dir, const char *name, struct stat *buf)
{
	int rc;

	if (HAVE_DIRFD)
	{
		int dfd;

		dfd = GATEG_DIRFD(dir);
		if (dfd < 0)
		{
			perror("dirfd");

			return -1;
		}
		rc = fstat(dfd, buf);
	}
	else
	{
		rc = stat(".", buf);
	}

	return rc;
}

/* ---------------------------- test functions ----------------------------- */

static int __test_evar1(void)
{
	gateg_cfg_t gg_cfg;
	int ret;
	int tablen;
	int i;
	gateg_evarhandler_t eh1[999] = {
		GATEG_VAR_KEEP("FOO4"),
		GATEG_VAR_SET("IFS", " \t\n"),
		GATEG_VAR_SET("PATH", GATEG__PATH_STDPATH),
		GATEG_VAR_REJECT_PATH("TERMCAP"),
		GATEG_VAR_REJECT_PATH("TERMCAP2"),
		GATEG_VAR_REJECT_PATH("TERMCAP3"),
		GATEG_VAR_REJECT_PATH("TERMCAP4"),
		GATEG_VAR_FILTER_LANG("LANG"),
		GATEG_VAR_FILTER_LANG("LANG2"),
		GATEG_VAR_FILTER_LANG("LANG3"),
		GATEG_VAR_FILTER_LANG("LANG4"),
		GATEG_VAR_FILTER_LANG("LANG5"),
		GATEG_VAR_FILTER_LANG("LANG6"),
		GATEG_VAR_FILTER_LANG("LANG7"),
		GATEG_VAR_FILTER_INT("LINES"),
		GATEG_VAR_FILTER_INT("LINES2"),
		GATEG_VAR_FILTER_INT("LINES3"),
		GATEG_VAR_FILTER_INT("LINES4"),
		GATEG_VAR_FILTER_TERM("TERM"),
		GATEG_VAR_FILTER_TERM("TERM2"),
		GATEG_VAR_FILTER_TERM("TERM3"),
		GATEG_VAR_FILTER_TERM("TERM4"),
		GATEG_VAR_SET("FOO2", "BAR"),
		GATEG_VAR_SET("FOO3", 0),
		GATEG_VAR_END_OF_TABLE
	};
	struct
	{
		char *var;
		char *val;
	} vars[] = {
		{ "IFS", "xyzzy" },
		{ "PATH", "foobar" },
		{ "TERMCAP", "foo/" },
		{ "TERMCAP2", "foo/bar" },
		{ "TERMCAP3", "/bar" },
		{ "TERMCAP4", "foobar" },
		{ "LANG", "foo%" },
		{ "LANG2", "foo%bar" },
		{ "LANG3", "%bar" },
		{ "LANG4", "foo/" },
		{ "LANG5", "foo/bar" },
		{ "LANG6", "/bar" },
		{ "LANG7", "La_ng" },
		{ "LINES", "999a999" },
		{ "LINES2", "a999" },
		{ "LINES3", "999a" },
		{ "LINES4", "9753108642" },
		{ "TERM", "aaa!aaa" },
		{ "TERM2", "aaa!" },
		{ "TERM3", "!aaa" },
		{ "TERM4", "abcdefghijklmnopqrstuvwxyz"
		  "ABCDEFGHIJKLMNOPQRSTUVWXYZ-/:+._0123456789" },
		{ "FOO", "BAR" },
		{ 0, 0 }
	};
	char *results[] = {
		"IFS= \t\n",
		"PATH="GATEG__PATH_STDPATH,
		"TERMCAP4=foobar",
		"LANG7=La_ng",
		"LINES4=9753108642",
		"TERM4=abcdefghijklmnopqrstuvwxyz"
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ-/:+._0123456789",
		"FOO2=BAR",
		0
	};

	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.evars_clean = GATEG_OP_DFLT_PARANOID;
	gg_cfg.evars_handlers = eh1;
	for (i = 0; vars[i].var != 0; i++)
	{
		char *s;
		int l;

		l = __gateg_strlen(vars[i].var) + 1 +
			__gateg_strlen(vars[i].val) + 1;
		s = (char *)malloc(l);
		sprintf(s, "%s=%s", vars[i].var, vars[i].val);
		putenv(s);
	}
	tablen = i;
	gateg_safe_init(&gg_cfg);
	for (i = 0, ret = RET_OK; environ[i] != 0; i++)
	{
		fprintf(stderr, "%d: %s\n", i, environ[i]);
		if (i == tablen)
		{
			fprintf(
				stderr,
				"-- resulting environment is too long\n");
			ret = RET_FAILED;
		}
		else if (results[i] == 0 ||
			 __str_equal(environ[i], results[i]) == 0)
		{
			fprintf(
				stderr, "-- expected\n   %s\n",
				(results[i] == 0) ? "(null)" : results[i]);
			ret = RET_FAILED;
		}
	}

	return ret;
}

static int __test_core_do_dump(void)
{
	*(int *)0 = 0;

	return RET_FAILED;
}

static int __test_core_no_dump(void)
{
	gateg_cfg_t gg_cfg;

	if (TEST_CORE == 1)
	{
		gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
		gg_cfg.ops.core_disable = GATEG_OP_DFLT_PARANOID;
		gateg_safe_init(&gg_cfg);
		*(int *)0 = 0;

		return RET_FAILED;
	}
	else
	{
		fprintf(stderr, "    setrlimit is not supported\n");
		return RET_SKIPPED;
	}
}

static int __test_delay(void)
{
	gateg_cfg_t gg_cfg;
	int rc;

	/* disable umask */
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.umask_set = GATEG_OP_SET | GATEG_OP_DELAY_2;
	gg_cfg.umask = 0077;
	rc = gateg_safe_init(&gg_cfg);
	if (rc != 1)
	{
		fprintf(stderr, "    delay 2 failed\n");
	}
	rc = gateg_safe_init(&gg_cfg);
	if (rc != 1)
	{
		fprintf(stderr, "    delay 1 failed\n");
	}
	rc = gateg_safe_init(&gg_cfg);
	if (rc != 0)
	{
		fprintf(stderr, "    delay 0 failed\n");
	}
	ASSERT_ABORT(gateg_safe_init(&gg_cfg));

	return RET_OK;
}

static int __test_umask(void)
{
	gateg_cfg_t gg_cfg;
	mode_t m;

	/* disable umask */
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.umask_set = GATEG_OP_SET;
	gg_cfg.umask = 0000;
	gateg_safe_init(&gg_cfg);
	m = umask(0000);
	if (m != 0000)
	{
		fprintf(stderr, "    umask 0000 failed (0%o)\n", (int)m);
		return RET_FAILED;
	}
	/* set umask */
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.umask_set = GATEG_OP_SET;
	gg_cfg.umask = 0070;
	gateg_safe_init(&gg_cfg);
	m = umask(0070);
	if (m != 0070)
	{
		fprintf(stderr, "    umask 0070 failed (0%o)\n", (int)m);
		return RET_FAILED;
	}
	/* paranoid umask */
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.umask_set = GATEG_OP_SET_PARANOID;
	gg_cfg.umask = 0700;
	gateg_safe_init(&gg_cfg);
	m = umask(0770);
	if (m != 0770)
	{
		fprintf(stderr, "    umask 0770 failed (0%o)\n", (int)m);
		return RET_FAILED;
	}

	return RET_OK;
}

static RETSIGTYPE __test_handler(int sig) { }

static int __test_sighandlers(void)
{
	void *handler1;
	void *handler2;
	gateg_cfg_t gg_cfg;
	int i;

	for (i = 0; i <= GATEG_NSIG; i++)
	{
		signal(i, __test_handler);
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.sighand_reset = GATEG_OP_SET;
	gateg_safe_init(&gg_cfg);
	for (i = 0; i <= GATEG_NSIG; i++)
	{
		handler1 = (void *)signal(i, SIG_DFL);
		if (handler1 == (void *)SIG_ERR)
		{
			continue;
		}
		handler2 = (void *)signal(i, SIG_DFL);
		if (handler1 != handler2)
		{
			fprintf(
				stderr,
				"    signal %d was not reset properly\n", i);
			return RET_FAILED;
		}
	}

	return RET_OK;
}

static int __test_sighandlers_sequence(void)
{
	gateg_cfg_t gg_cfg;

	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.sighand_reset = GATEG_OP_SET | GATEG_OP_DELAY_1;
	gg_cfg.ops.sighand_verify = GATEG_OP_SET;
	ASSERT_ABORT(gateg_safe_init(&gg_cfg));

	return RET_OK;
}

static int __test_stdfds1(void)
{
	gateg_cfg_t gg_cfg;

	fclose(stdin);
	close(0);
	fclose(stdout);
	close(1);
	fclose(stderr);
	close(2);
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.fds_std = GATEG_OP_DFLT;
	gateg_safe_init(&gg_cfg);
	if (fileno(stdin) != 0)
	{
		return RET_FAILED;
	}
	if (fileno(stdout) != 1)
	{
		return RET_FAILED;
	}
	if (fileno(stderr) != 2)
	{
		return RET_FAILED;
	}

	return RET_OK;
}

static int __test_stdfds2(void)
{
	gateg_cfg_t gg_cfg;
	FILE *f;

	fclose(stdin);
	close(0);
	fclose(stdout);
	close(1);
	fclose(stderr);
	close(2);
	f = freopen(GATEG__PATH_DEVNULL, "rb", stdin);
	if (f == 0 || fileno(f) != 0)
	{
		return RET_FAILED;
	}
	f = freopen(GATEG__PATH_DEVNULL, "wb", stdout);
	if (f == 0 || fileno(f) != 1)
	{
		return RET_FAILED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.fds_std = GATEG_OP_DFLT;
	gateg_safe_init(&gg_cfg);
	if (fileno(stdin) != 0)
	{
		return RET_FAILED;
	}
	if (fileno(stdout) != 1)
	{
		return RET_FAILED;
	}
	if (fileno(stderr) != 2)
	{
		return RET_FAILED;
	}

	return RET_OK;
}

static int __test_stdfds3(void)
{
	gateg_cfg_t gg_cfg;
	FILE *f;

	fclose(stdin);
	close(0);
	fclose(stdout);
	close(1);
	fclose(stderr);
	close(2);
	f = fopen(GATEG__PATH_DEVNULL, "rb");
	if (f == 0 || fileno(f) != 0)
	{
		return RET_FAILED;
	}
	f = fopen(GATEG__PATH_DEVNULL, "wb");
	if (f == 0 || fileno(f) != 1)
	{
		return RET_FAILED;
	}
	f = fopen(GATEG__PATH_DEVNULL, "wb");
	if (f == 0 || fileno(f) != 2)
	{
		return RET_FAILED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.fds_std = GATEG_OP_DFLT_PARANOID;
	gateg_safe_init(&gg_cfg);
	if (fileno(stdin) != 0)
	{
		return RET_FAILED;
	}
	if (fileno(stdout) != 1)
	{
		return RET_FAILED;
	}
	if (fileno(stderr) != 2)
	{
		return RET_FAILED;
	}

	return RET_OK;
}

static int __test_fds(void)
{
	gateg_cfg_t gg_cfg;
	struct stat buf;
	DIR *d;
	int fd;
	int max;
	int rc;
	int i;

	/* open some files and directories */
	for (max = 0, i = 0; ; i++)
	{
		if (i % 2 == 0)
		{
			fd = open(GATEG__PATH_DEVNULL, O_RDONLY);
		}
		else
		{
			d = opendir(".");
			if (d == 0)
			{
				break;
			}
			fd = GATEG_DIRFD(d);
		}
		if (fd < 0)
		{
			break;
		}
		if (max < fd)
		{
			max = fd;
		}
	}
	max++;
	fprintf(stderr, "    opened %d descriptors\n", i);
	if (i < 2)
	{
		fprintf(stderr, "    could not open enough files for test\n");

		return RET_FAILED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.fds_close = GATEG_OP_SET;
	gateg_safe_init(&gg_cfg);
	for (i = 3; i <= max; i++)
	{
		rc = fstat(i, &buf);
		if (rc == 0)
		{
			fprintf(
				stderr, "    descriptor %d was not closed\n",
				i);

			return RET_FAILED;
		}
		if (errno != EBADF)
		{
			perror("fstat");

			return RET_FAILED;
		}
	}

	return RET_OK;
}

static int __test_cwd(void)
{
	gateg_cfg_t gg_cfg;
	struct stat obuf;
	struct stat nbuf;
	DIR *odir;
	DIR *ndir;
	int rc;

	odir = opendir(".");
	if (odir == 0)
	{
		perror("opendir");

		return RET_FAILED;
	}
	rc = __pstat(odir, ".", &obuf);
	if (rc != 0)
	{
		perror("(f)stat");

		return RET_FAILED;
	}
	rc = chdir("..");
	if (rc != 0)
	{
		perror("chdir");

		return RET_FAILED;
	}
	ndir = opendir(".");
	if (ndir == 0)
	{
		perror("opendir");

		return RET_FAILED;
	}
	rc = __pstat(ndir, ".", &nbuf);
	if (rc != 0)
	{
		perror("(f)stat");

		return RET_FAILED;
	}
	if (obuf.st_dev == nbuf.st_dev && obuf.st_ino == nbuf.st_ino)
	{
		/* still in the same directory */
		fprintf(stderr, "    chdir stayed in the same directory\n");

		return RET_FAILED;
	}
	rc = closedir(ndir);
	if (rc != 0)
	{
		perror("closedir");

		return RET_FAILED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.cwd_set = GATEG_OP_SET;
	gg_cfg.cwd = TEST_DIR;
	gateg_safe_init(&gg_cfg);
	ndir = opendir(".");
	if (ndir == 0)
	{
		perror("opendir");

		return RET_FAILED;
	}
	rc = __pstat(ndir, ".", &nbuf);
	if (rc != 0)
	{
		perror("fstat");

		return RET_FAILED;
	}
	if (obuf.st_dev != nbuf.st_dev || obuf.st_ino != nbuf.st_ino)
	{
		/* we're somewhere else */
		return RET_FAILED;
	}
	closedir(ndir);
	closedir(odir);

	return RET_OK;
}

static int __test_cwd2(void)
{
	gateg_cfg_t gg_cfg;
	DIR *dir;

	dir = opendir(NONEXIST_DIR);
	if (dir != 0)
	{
		fprintf(stderr, "    could enter " NONEXIST_DIR "\n");
		closedir(dir);
		return RET_FAILED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.cwd_set = GATEG_OP_SET;
	gg_cfg.cwd = NONEXIST_DIR;
	ASSERT_ABORT(gateg_safe_init(&gg_cfg));

	return RET_OK;
}

static int __test_set_uid(void)
{
	gateg_cfg_t gg_cfg;

	if (geteuid() != 0)
	{
		fprintf(stderr, "    this test can only run with uid 0\n");
		return RET_SKIPPED;
	}
	if (getegid() != 0)
	{
		fprintf(stderr, "    this test can only run with gid 0\n");
		return RET_SKIPPED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.priv_set = GATEG_OP_SET;
	gg_cfg.priv_uid = 1;
	gg_cfg.priv_gid = 1;
	gateg_safe_init(&gg_cfg);
	if (geteuid() != 1)
	{
		fprintf(stderr, "    failed to set euid\n");
		return RET_FAILED;
	}
	if (getegid() != 1)
	{
		fprintf(stderr, "    failed to set egid\n");
		return RET_FAILED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.priv_set = GATEG_OP_SET;
	gg_cfg.priv_uid = 0;
	gg_cfg.priv_gid = 0;
	gateg_safe_init(&gg_cfg);
	if (geteuid() != 0)
	{
		fprintf(stderr, "    failed to restore euid\n");
		return RET_FAILED;
	}
	if (getegid() != 0)
	{
		fprintf(stderr, "    failed to restore egid\n");
		return RET_FAILED;
	}

	return RET_OK;
}

static int __test_drop_uid(void)
{
	gateg_cfg_t gg_cfg;

	if (geteuid() != 0)
	{
		fprintf(stderr, "    this test can only run with uid 0\n");
		return RET_SKIPPED;
	}
	if (getegid() != 0)
	{
		fprintf(stderr, "    this test can only run with gid 0\n");
		return RET_SKIPPED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.priv_set = GATEG_OP_SET;
	gg_cfg.ops.priv_drop = GATEG_OP_SET;
	gg_cfg.priv_uid = 1;
	gg_cfg.priv_gid = 1;
	gateg_safe_init(&gg_cfg);
	if (getuid() != 1)
	{
		fprintf(stderr, "    failed to set uid\n");
		return RET_FAILED;
	}
	if (geteuid() != 1)
	{
		fprintf(stderr, "    failed to set euid\n");
		return RET_FAILED;
	}
	if (getgid() != 1)
	{
		fprintf(stderr, "    failed to set gid\n");
		return RET_FAILED;
	}
	if (getegid() != 1)
	{
		fprintf(stderr, "    failed to set egid\n");
		return RET_FAILED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.priv_set = GATEG_OP_SET;
	gg_cfg.ops.priv_drop = GATEG_OP_SET;
	gg_cfg.priv_uid = 0;
	gg_cfg.priv_gid = 0;
	gateg_safe_init(&gg_cfg);
	if (getuid() != 1)
	{
		fprintf(stderr, "    could restore uid\n");
		return RET_FAILED;
	}
	if (geteuid() != 1)
	{
		fprintf(stderr, "    could restore euid\n");
		return RET_FAILED;
	}
	if (getgid() != 1)
	{
		fprintf(stderr, "    could restore gid\n");
		return RET_FAILED;
	}
	if (getegid() != 1)
	{
		fprintf(stderr, "    could restore egid\n");
		return RET_FAILED;
	}

	return RET_OK;
}

static int __test_chroot(void)
{
	gateg_cfg_t gg_cfg;
	struct stat buf;
	struct stat buf2;
	DIR *dir;
	DIR *dir2;
	int rc;

	if (geteuid() != 0)
	{
		fprintf(stderr, "    this test can only run with uid 0\n");
		return RET_SKIPPED;
	}
	if (getegid() != 0)
	{
		fprintf(stderr, "    this test can only run with gid 0\n");
		return RET_SKIPPED;
	}
	dir = opendir(".");
	if (dir == 0)
	{
		perror("opendir");
		return RET_FAILED;
	}
	rc = __pstat(dir, ".", &buf);
	if (rc != 0)
	{
		perror("fstat");
		return RET_FAILED;
	}
	rc = chdir("..");
	if (rc != 0)
	{
		perror("chdir");
		return RET_FAILED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.priv_set = GATEG_OP_SET;
	gg_cfg.ops.priv_drop = GATEG_OP_SET;
	gg_cfg.ops.fds_close = GATEG_OP_SET;
	gg_cfg.ops.chroot_call = GATEG_OP_SET;
	gg_cfg.ops.chroot_verify = GATEG_OP_SET;
	gg_cfg.priv_uid = 1;
	gg_cfg.priv_gid = 1;
	gg_cfg.chroot_dir = TEST_DIR;
	gateg_safe_init(&gg_cfg);
	dir2 = opendir(".");
	if (dir2 == 0)
	{
		perror("opendir");
		return RET_FAILED;
	}
	rc = __pstat(dir2, ".", &buf2);
	if (rc != 0)
	{
		perror("fstat");
		return RET_FAILED;
	}
	if (buf.st_dev != buf2.st_dev || buf.st_ino != buf2.st_ino)
	{
		fprintf(stderr, "    wrong chroot directory\n");
		return RET_FAILED;
	}

	return RET_OK;
}

static int __test_chroot_still_root(void)
{
	gateg_cfg_t gg_cfg;
	int rc;

	if (geteuid() != 0)
	{
		fprintf(stderr, "    this test can only run with uid 0\n");
		return RET_SKIPPED;
	}
	if (getegid() != 0)
	{
		fprintf(stderr, "    this test can only run with gid 0\n");
		return RET_SKIPPED;
	}
	rc = chdir("..");
	if (rc != 0)
	{
		perror("chdir");
		return RET_FAILED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.fds_close = GATEG_OP_SET;
	gg_cfg.ops.chroot_call = GATEG_OP_SET;
	gg_cfg.ops.chroot_verify = GATEG_OP_SET;
	gg_cfg.chroot_dir = TEST_DIR;
	ASSERT_ABORT(gateg_safe_init(&gg_cfg));

	return RET_OK;
}

static int __test_chroot_not_root(void)
{
	gateg_cfg_t gg_cfg;
	int rc;

	if (geteuid() != 0)
	{
		fprintf(stderr, "    this test can only run with uid 0\n");
		return RET_SKIPPED;
	}
	if (getegid() != 0)
	{
		fprintf(stderr, "    this test can only run with gid 0\n");
		return RET_SKIPPED;
	}
	rc = chdir("..");
	if (rc != 0)
	{
		perror("chdir");
		return RET_FAILED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.priv_set = GATEG_OP_SET;
	gg_cfg.ops.priv_drop = GATEG_OP_SET;
	gg_cfg.priv_uid = 1;
	gg_cfg.priv_gid = 1;
	gateg_safe_init(&gg_cfg);
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.chroot_call = GATEG_OP_SET;
	gg_cfg.chroot_dir = TEST_DIR;
	ASSERT_ABORT(gateg_safe_init(&gg_cfg));

	return RET_OK;
}

static int __test_chroot_invaliddir(void)
{
	gateg_cfg_t gg_cfg;
	int rc;

	if (geteuid() != 0)
	{
		fprintf(stderr, "    this test can only run with uid 0\n");
		return RET_SKIPPED;
	}
	if (getegid() != 0)
	{
		fprintf(stderr, "    this test can only run with gid 0\n");
		return RET_SKIPPED;
	}
	rc = chdir("..");
	if (rc != 0)
	{
		perror("chdir");
		return RET_FAILED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.priv_set = GATEG_OP_SET;
	gg_cfg.ops.priv_drop = GATEG_OP_SET;
	gg_cfg.ops.fds_close = GATEG_OP_SET;
	gg_cfg.ops.chroot_call = GATEG_OP_SET;
	gg_cfg.ops.chroot_verify = GATEG_OP_SET;
	gg_cfg.priv_uid = 1;
	gg_cfg.priv_gid = 1;
	gg_cfg.chroot_dir = NONEXIST_DIR;
	gateg_safe_init(&gg_cfg);
	ASSERT_ABORT(gateg_safe_init(&gg_cfg));

	return RET_OK;
}

static int __test_chroot_opendir(void)
{
	gateg_cfg_t gg_cfg;
	DIR *dir;
	int rc;

	if (geteuid() != 0)
	{
		fprintf(stderr, "    this test can only run with uid 0\n");
		return RET_SKIPPED;
	}
	if (getegid() != 0)
	{
		fprintf(stderr, "    this test can only run with gid 0\n");
		return RET_SKIPPED;
	}
	dir = opendir(".");
	if (dir == 0)
	{
		perror("opendir");
		return RET_FAILED;
	}
	rc = chdir("..");
	if (rc != 0)
	{
		perror("chdir");
		return RET_FAILED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.priv_set = GATEG_OP_SET;
	gg_cfg.ops.priv_drop = GATEG_OP_SET;
	gg_cfg.ops.chroot_call = GATEG_OP_SET;
	gg_cfg.ops.chroot_verify = GATEG_OP_SET;
	gg_cfg.priv_uid = 1;
	gg_cfg.priv_gid = 1;
	gg_cfg.chroot_dir = TEST_DIR;
	ASSERT_ABORT(gateg_safe_init(&gg_cfg));

	return RET_OK;
}

static int __test_chroot_sequence(void)
{
	gateg_cfg_t gg_cfg;
	int rc;

	if (geteuid() != 0)
	{
		fprintf(stderr, "    this test can only run with uid 0\n");
		return RET_SKIPPED;
	}
	if (getegid() != 0)
	{
		fprintf(stderr, "    this test can only run with gid 0\n");
		return RET_SKIPPED;
	}
	rc = chdir("..");
	if (rc != 0)
	{
		perror("chdir");
		return RET_FAILED;
	}
	gateg_init_cfg(&gg_cfg, GATEG_INIT_NOP);
	gg_cfg.ops.priv_set = GATEG_OP_SET;
	gg_cfg.ops.priv_drop = GATEG_OP_SET;
	gg_cfg.ops.fds_close = GATEG_OP_SET;
	gg_cfg.ops.chroot_call = GATEG_OP_SET | GATEG_OP_DELAY_1;
	gg_cfg.ops.chroot_verify = GATEG_OP_SET;
	gg_cfg.priv_uid = 1;
	gg_cfg.priv_gid = 1;
	gg_cfg.chroot_dir = TEST_DIR;
	ASSERT_ABORT(gateg_safe_init(&gg_cfg));

	return RET_OK;
}

static int __test_run_command1(void)
{
	int rc;
	char *argsf[] = { "/bin/false", 0 };
	char *argst[] = { "/bin/true", 0 };

	rc = gateg_safe_run_command(argst);
	if (rc != 0)
	{
		return RET_FAILED;
	}
	rc = gateg_safe_run_command(argsf);
	if (rc != 1)
	{
		return RET_FAILED;
	}

	return RET_OK;
}

static int __test_run_command2(void)
{
	int rc;
	char *argsf[] = { "/bin/false", 0 };
	char *argst[] = { "/bin/true", 0 };
	gateg_cfg_t cfg;

	gateg_init_cfg(&cfg, GATEG_INIT_NOP);
	rc = gateg_safe_run_command_with_cfg(&cfg, argst);
	if (rc != 0)
	{
		return RET_FAILED;
	}
	rc = gateg_safe_run_command_with_cfg(&cfg, argsf);
	if (rc != 1)
	{
		return RET_FAILED;
	}

	return RET_OK;
}

/* ---------------------------- local variables ---------------------------- */

static test_table_entry_t test_table[] =
{
	{ __test_delay, "delay", 0, RET_OK },
	{ __test_umask, "umask", 0, RET_OK },
	{ __test_sighandlers, "sighandlers", "reset signal handlers", RET_OK },
	{ __test_sighandlers_sequence, "sighandlers2", "verify before reset",
	  RET_OK },
	{ __test_evar1, "evar1", "test environment variable filters", RET_OK },
	{ __test_stdfds1, "stdfds1", "reopen closed std descriptors", RET_OK },
	{ __test_stdfds2, "stdfds2", "check reopen std descriptors", RET_OK },
	{ __test_stdfds3, "stdfds3", "force reopen std descriptors", RET_OK },
	{ __test_fds, "fds", "close all descriptors", RET_OK },
	{ __test_cwd, "cwd", "change working directory", RET_OK },
	{ __test_cwd2, "cwd2", "invalid working directory", RET_OK },
	{ __test_chroot, "chroot1", "working chroot", RET_OK },
	{ __test_chroot_invaliddir, "chroot2", "chroot with invalid dir",
	  RET_OK },
	{ __test_chroot_opendir, "chroot3", "chroot with open dir", RET_OK },
	{ __test_chroot_still_root, "chroot4", "root in the chroot", RET_OK },
	{ __test_chroot_not_root, "chroot5", "chroot as normal user", RET_OK },
	{ __test_chroot_sequence, "chroot6", "verify before chroot", RET_OK },

	{ __test_set_uid, "setuid", "set the uid/gid and switch back",
	  RET_OK },
	{ __test_drop_uid, "dropuid", "drop the uid/gid", RET_OK },
	{ __test_core_do_dump, "core1", "dump a core file", RET_EXPECT_CORE },
	{ __test_core_no_dump, "core2", "do not dump a core file",
	  RET_EXPECT_NO_CORE },
	{ __test_run_command1, "runcmd1", "run an external command" },
	{ __test_run_command2, "runcmd2", "run an external command" },
	{ 0, 0, 0, 0 }
};

/* ---------------------------- local functions ---------------------------- */

static int __check_core_file(int mode, const char *core_name)
{
	int expect_core;
	int have_core;
	struct stat buf;
	int rc;

	rc = stat(core_name, &buf);
	if (rc != 0)
	{
		if (errno != ENOENT)
		{
			perror("stat");

			return RET_FAILED;
		}
		have_core = 0;
	}
	else if (S_ISREG(buf.st_mode))
	{
		have_core = 1;
	}
	else
	{
		fprintf(stderr, "%s is not a regular file\n", core_name);
		return RET_FAILED;
	}
	if (have_core)
	{
		rc = unlink(core_name);
		if (rc != 0)
		{
			perror("unlink");

			return RET_FAILED;
		}
	}
	expect_core = (mode == RET_EXPECT_CORE);
	if (expect_core == 1 && have_core == 0)
	{
		return RET_SKIP_NO_CORE;
	}
	if (expect_core == 0 && have_core == 1)
	{
		return RET_FAILED;
	}

	return RET_OK;
}

static int __run_test(test_table_entry_t *test)
{
	pid_t child = 0;
	pid_t pid = 0;
	int status = 0;
	int rc;
	static int do_skip_no_core = 0;

	child = 0;
	pid = 0;
	status = 0;
	fprintf(stderr, "*** executing test %s\n", test->t_name);
	if (test->t_msg != 0)
	{
		fprintf(stderr, "    %s\n", test->t_msg);
	}
	if (GATEG_TEST_DEBUG == 0)
	{
		child = fork();
		switch (child)
		{
		case (pid_t)-1:
			perror("fork");
			abort();
		case 0:
			/* child */
			rc = test->t_func();
			exit(rc);
		default:
			/* parent */
			pid = wait(&status);
			rc = WEXITSTATUS(status);
			break;
		}
	}
	else
	{
		rc = test->t_func();
	}
	switch (test->t_resulttype)
	{
	case RET_EXPECT_CORE:
	case RET_EXPECT_NO_CORE:
		if (rc != RET_SKIPPED)
		{
			rc = __check_core_file(test->t_resulttype, CORE_NAME);
		}
		if (rc == RET_SKIP_NO_CORE)
		{
			rc = __check_core_file(test->t_resulttype, CORE_NAME2);
		}
		if (rc == RET_SKIP_NO_CORE)
		{
			rc = __check_core_file(test->t_resulttype, CORE_NAME3);
		}
		if (rc == RET_SKIP_NO_CORE)
		{
			char name[128];

			sprintf(name, "%s.%d", CORE_NAME4_PREFIX, (int)pid);
			rc = __check_core_file(test->t_resulttype, name);
		}
		if (rc == RET_SKIP_NO_CORE)
		{
			char name[128];

			sprintf(name, "%s.%d", CORE_NAME5_PREFIX, (int)pid);
			rc = __check_core_file(test->t_resulttype, name);
		}
		if (do_skip_no_core == 1 &&
		    test->t_resulttype == RET_EXPECT_NO_CORE)
		{
			rc = RET_SKIPPED;
		}
		break;
	default:
		break;
	}
	switch (rc)
	{
	case RET_OK:
		fprintf(stderr, "OK\n");
		break;
	case RET_SKIPPED:
		fprintf(stderr, "SKIPPED test\n");
		break;
	case RET_SKIP_NO_CORE:
		fprintf(stderr, "OK, machine does not write core files\n");
		do_skip_no_core = 1;
		break;
	case RET_FAILED:
	default:
		fprintf(stderr, "FAILED with return code %d\n", rc);
		break;
	}

	return rc;
}

static void __setup_test_env(void)
{
	int rc;

	rmdir(TEST_DIR);
	umask(0022);
	rc = mkdir(TEST_DIR, 0755);
	if (rc != 0)
	{
		perror("mkdir");
		fprintf(stderr, "failed to create directory '%s'\n", TEST_DIR);
		fprintf(
			stderr,
			"if it exists and is not empty, please remove it"
			" manually\n");
		exit(2);
	}
	rc = chdir(TEST_DIR);
	if (rc != 0)
	{
		perror("chdir");
		exit(2);
	}

	return;
}

static void __cleanup_test_env(void)
{
	chdir("..");
	rmdir(TEST_DIR);

	return;
}

/* ---------------------------- interface functions ------------------------ */

int main(void)
{
	int tests_failed;
	int tests_skipped;
	int rc;
	int i;

	fprintf(
		stderr, "If this program aborts, there is a flaw in the"
		" tested code\n"
		"(Aborted sub-processes are part of the tests.)\n");
	__setup_test_env();
	tests_failed = 0;
	tests_skipped = 0;
	for (i = 0; test_table[i].t_func != 0; i++)
	{
		rc = __run_test(&test_table[i]);
		switch (rc)
		{
		case RET_FAILED:
			tests_failed++;
			break;
		case RET_SKIPPED:
		case RET_SKIP_NO_CORE:
			tests_skipped++;
			break;
		case RET_OK:
			break;
		default:
			fprintf(
				stderr, "\nERROR: invalid return code %d\n",
				rc);
			abort();
			exit(1);
			break;
		}
	}
	__cleanup_test_env();
	if (tests_skipped != 0)
	{
		fprintf(
			stderr, "\nWARNING: %d tests have been skipped\n",
			tests_skipped);
	}
	if (tests_failed != 0)
	{
		fprintf(stderr, "\nERROR: at least one test has failed\n");

		return 1;
	}

	return 0;
}
